#!/usr/bin/perl
#_*****************************************************************************
#_*                                                                           *
#_* File name: quma.pl                                                        *
#_* Author: Yuichi Kumaki (kuma@cdb.riken.go.jp or yuichi@kumaki.jp)          *
#_* Copyright (C) 2010 Yuichi Kumaki                                          *
#_* Laboratory for Mammalian Epigenetic Studies,                              *
#_* Center for Developmental Biology, RIKEN                                   *
#_* 2-2-3 Minatojima-Minamimachi, Chuo-ku, Kobe 650-0047, Japan               *
#_*                                                                           *
#_* 2010/10/22 First version                                                  *
#_*                                                                           *
#_*****************************************************************************
#_*                                                                           *
#_* This program is free software: you can redistribute it and/or modify      *
#_* it under the terms of the GNU General Public License as published by      *
#_* the Free Software Foundation, either version 3 of the License, or         *
#_* (at your option) any later version.                                       *
#_*                                                                           *
#_* This program is distributed in the hope that it will be useful,           *
#_* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
#_* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
#_* GNU General Public License for more details.                              *
#_*                                                                           *
#_* You should have received a copy of the GNU General Public License         *
#_* along with this program.  If not, see <http://www.gnu.org/licenses/>.     *
#_*                                                                           *
#_*****************************************************************************
############ Please change these values for your environment! ##################
                                                                               #
###  path of the needle program                                                #
use constant NEEDLE   => 'needle';                                             #
                                                                               #
###  full path of the temporary file directory                                 #
###  (necessary to writable)                                                   #
use constant TEMPDIR  => '/tmp/';                                              #
                                                                               #
################################################################################

use 5.6.0;
use strict;
use warnings;
use Statistics::Lite qw(:funcs);

sub usage ();
sub makeTime (;$);
sub curateSeq ($);
sub parseSeq ($);
sub parseGenome ($);
sub parseBiseq ($);
sub parseMulti ($);
sub multiFastaParse ($);
sub revComp ($);
sub fastaMake ($$;$);
sub fastaPrint ($$$;$$);
sub execNeedle ($$$);

use constant UNCONVL  => 5;
use constant PCONVL   => 95.0;
use constant MISL     => 10;
use constant PERCL    => 90.0;
use constant MAT  =>
'    A   T   G   C   S   W   R   Y   K   M   B   V   H   D   N   U
A   5  -4  -4  -4  -4   1   1  -4  -4   1  -4  -1  -1  -1  -2  -4
T  -4   5  -4   5  -4   1  -4   1   1  -4  -1  -4  -1  -1  -2   5
G  -4  -4   5  -4   1  -4   1  -4   1  -4  -1  -1  -4  -1  -2  -4
C  -4  -4  -4   5   1  -4  -4   1  -4   1  -1  -1  -1  -4  -2  -4
S  -4  -4   1   1  -1  -4  -2  -2  -2  -2  -1  -1  -3  -3  -1  -4
W   1   1  -4  -4  -4  -1  -2  -2  -2  -2  -3  -3  -1  -1  -1   1
R   1  -4   1  -4  -2  -2  -1  -4  -2  -2  -3  -1  -3  -1  -1  -4
Y  -4   1  -4   1  -2  -2  -4  -1  -2  -2  -1  -3  -1  -3  -1   1
K  -4   1   1  -4  -2  -2  -2  -2  -1  -4  -1  -3  -3  -1  -1   1
M   1  -4  -4   1  -2  -2  -2  -2  -4  -1  -3  -1  -1  -3  -1  -4
B  -4  -1  -1  -1  -1  -3  -3  -1  -1  -3  -1  -2  -2  -2  -1  -1
V  -1  -4  -1  -1  -1  -3  -1  -3  -3  -1  -2  -1  -2  -2  -1  -4
H  -1  -1  -4  -1  -3  -1  -3  -1  -3  -1  -2  -2  -1  -2  -1  -1
D  -1  -1  -1  -4  -3  -1  -1  -3  -1  -3  -2  -2  -2  -1  -1  -1
N  -2  -2  -2  -2  -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  -2
U  -4   5  -4  -4  -4   1  -4   1   1  -4  -1  -4  -1  -1  -2   5
';
use constant CPGMAT   => TEMPDIR . '/EDNA_CPG';
use constant NDLOPT   => ' -stdout -auto -gapopen 10.0 -gapextend 0.5 ' .
                         '-aformat markx3 -datafile ' . CPGMAT . '|';
use constant OUTFORM  => qq[Summary of information
Length of target genome sequence\t\%d
Number of CpGs\t\%d
Number of bisulfite sequences (used)\t\%d
Number of bisulfite sequences (excluded)\t\%d
Number of bisulfite sequences (total)\t\%d

Condition to exclude low quality sequences
Upper limit of unconverted CpHs\t\%d
Lower limit of percent converted CpHs\t\%f\%\%
Upper limit of alignment mismatches\t\%d
Lower limit of percent identity\t\%f\%\%

Methylation status of each CpG site
CpG position\t\%s
Number of methylated CpGs\t\%s
Number of CpGs\t\%s
ratio of methylated (\%\%)\t\%s
S.D. of \%\% methylatied (between CpGs)\t\%.1f
S.E. of \%\% methylatied (between CpGs)\t\%.2f
S.D. of \%\% methylatied (between sequences)\t\%.1f
S.E. of \%\% methylatied (between sequences)\t\%.2f

Bisulfite sequence information
No.\tsequence name\tmismatches(include gaps)\tgaps\talignment length\tpercent identity\tmethylated CpGs\tpercent methylated CpGs\tunconverted CpHs\tnumber of CpHs\tpercent converted CpHs\tmethylation pattern (U: unmethylated, M: methylated, A,C,G,T,N: mismatch, -: gap)
\%s
];
use constant ALI  => q{========================================================================
                     Alignment results No.%3d
========================================================================
--------------------------
Summary of information
--------------------------
Bisulfite sequence name                    %s
Conversion                                 %s => %s conversion
(conversion of %s strand of the genomic sequence)
Length of bisulfite sequence               %d
Length of target genome sequence           %d
Aligned region of bisulfite sequence       %d - %d
Alignment direction                        %s
Number of CpGs                             %d
Number of methylated CpGs                  %d
Percent methylated                         %.1f %%
Number of unconverted CpHs (CpA/CpT/CpC)   %d
Number of CpHs (CpA/CpT/CpC)               %d
Percent converted CpHs (CpA/CpT/CpC)       %.1f %%
Number of mismatches (include gaps)        %d
Number of gaps                             %d
Alignment length                           %d
Percent identity                           %.1f %%
Methylation pattern                        %s
(U: unmethylated, M: methylated, A,C,G,T,N: mismatch, -: gap)

--------------------------
Target genome sequence
--------------------------
%s
--------------------------
Bisulfite sequence
--------------------------
%s
--------------------------
Alignment
--------------------------
'*' : Methylated CpG
'#' : Unmethylated CpG
'!' : Unconverted CpH
':' : Converted CpH
'|' : Match
' ' : Mismatch
--------------------------
%s
};
use constant LINE     => 60;

our (@ARGV  );
my ($time   );
my ($id     );
my ($arg    );
my ($gfile  );
my ($qfile  );
my ($unc    );
my ($pcon   );
my ($mis    );
my ($perc   );
my ($form   );
my ($file   );
my ($conv   );
my ($gseq   );
my (@qseq   );
my (@pos    );
my ($cpgf   );
my ($cpgr   );
my ($pos    );
my ($fh     );
my ($len    );
my ($qfileF );
my ($qfileR );
my ($gfileF );
my ($gfileR );
my ($qfilepF);
my ($qfilepR);
my ($gfilepF);
my ($gfilepR);
my ($fa     );
my ($ffres  );
my ($frres  );
my ($fres   );
my ($fdir   );
my ($rfres  );
my ($rrres  );
my ($rres   );
my ($rdir   );
my ($res    );
my ($gdir   );
my ($dir    );
my (@data   );
my ($out    );
my ($cpg    );
my ($cpgp   );
my ($ref    );
my ($cpgn   );
my ($i      );
my ($j      );
my (%meth   );
my ($allM   );
my ($allC   );
my ($numT   );
my ($numU   );
my ($numE   );
my ($meth   );
my ($pat    );
my ($m      );
my (@dsds   );
my (@dsdc   );
my ($sds    );
my ($sdc    );
my ($ses    );
my ($sec    );
my ($val1   );
my ($val2   );
my ($val3   );
my ($val4   );
my (@mali   );
my ($start  );
my ($stop   );
my ($g      );
my ($q      );
my ($k      );
my ($alen   );
my ($gen    );
my ($lc     );
my (%cpg    );
my ($qstart );
my ($qstop  );
my ($qlen   );
my ($output );
my ($bbase  );
my ($abase  );
my ($l      );
my ($gout   );
my ($ali    );
my ($qout   );
my ($cph    );
my ($cpgh   );
local ($_   );
local ($!   );
local ($a   );
local ($b   );

$time = makeTime;
$id = sprintf "$time%06d", $$;

while ($arg = shift) {
    if ($arg =~ /^-$/) {
	$file = '-';
    } elsif ($arg =~ /^-g$/) {
	$gfile = shift;
	usage () if ($gfile =~ /^-/);
    } elsif ($arg =~ /^-q$/) {
	$qfile = shift;
	usage () if ($qfile =~ /^-/);
    } elsif ($arg =~ /^-d$/) {
	$conv = shift;
	usage () if ($conv =~ /^-/);
    } elsif ($arg =~ /^-u$/) {
	$unc = shift;
	usage () if ($unc =~ /^-/);
    } elsif ($arg =~ /^-c$/) {
	$pcon = shift;
	usage () if ($pcon =~ /^-/);
    } elsif ($arg =~ /^-m$/) {
	$mis = shift;
	usage () if ($mis =~ /^-/);
    } elsif ($arg =~ /^-p$/) {
	$perc = shift;
	usage () if ($perc =~ /^-/);
    } elsif ($arg =~ /^-f$/) {
	$form = shift;
	usage () if ($form =~ /^-/);
    } elsif ($arg =~ /^-/ || $arg =~ /help/) {
	usage ();
    } else {
	usage () if ($file);
	$file = $arg;
    }
}

$conv ||= 0;
$unc  ||= UNCONVL;
$pcon ||= PCONVL;
$mis  ||= MISL;
$perc ||= PERCL;

if ($file) {
    usage () if ($gfile || $qfile);
} elsif ($gfile && $qfile) {
    usage () if ($file);
} else {
    usage ();
}

if ($file) {
    ($gseq, @qseq) = parseMulti ($file);
} else {
    $gseq = parseGenome ($gfile);
    @qseq = parseBiseq ($qfile);
}

$pos  = -1;
$cpgf = {};
$cpgr = {};
$len = length $gseq;

while (($pos = index $gseq, 'CG', ++$pos) != -1) {
    push @pos, $pos;
    $cpgf->{$pos} = 1;
    $cpgr->{$len - $pos - 2} = 1;
}

open $fh, '>' .CPGMAT or die "$!:".CPGMAT;
print $fh MAT;
close $fh;

$qfileF  = "que${id}F";
$qfileR  = "que${id}R";
$gfileF  = "genome${id}F";
$gfileR  = "genome${id}R";
$qfilepF = TEMPDIR . $qfileF;
$qfilepR = TEMPDIR . $qfileR;
$gfilepF = TEMPDIR . $gfileF;
$gfilepR = TEMPDIR . $gfileR;

fastaPrint $gseq, "$gfileF", $gfilepF;
fastaPrint revComp $gseq, "$gfileR", $gfilepR;

$pos = 0;
foreach $fa (@qseq) {

    ++$pos;
    $fa->{pos} = $pos;

    fastaPrint $fa->{seq}, "$qfileF", $qfilepF;
    fastaPrint revComp $fa->{seq}, "$qfileR", $qfilepR;

    unless ($conv == 1) {
	$ffres = execNeedle ($qfilepF, $gfilepF, $cpgf);
	$frres = execNeedle ($qfilepR, $gfilepF, $cpgf);

	{
	    if ($ffres->{'aliMis'} > $frres->{'aliMis'}) {
		$fres = $frres;
		$fdir = -1;
		last;
	    }

	    if ($ffres->{'aliMis'} < $frres->{'aliMis'}) {
		$fres = $ffres;
		$fdir = 1;
		last;
	    }

	    if ($ffres->{perc} > $frres->{perc}) {
		$fres = $ffres;
		$fdir = 1;
		last;
	    }

	    if ($ffres->{perc} < $frres->{perc}) {
		$fres = $frres;
		$fdir = -1;
		last;
	    }

	    if ($ffres->{unconv} > $frres->{unconv}) {
		$fres = $frres;
		$fdir = -1;
		last;
	    }

	    if ($ffres->{unconv} < $frres->{unconv}) {
		$fres = $ffres;
		$fdir = 1;
		last;
	    }

	    if ($ffres->{pconv} < $frres->{pconv}) {
		$fres = $frres;
		$fdir = -1;
		last;
	    }

	    if ($ffres->{pconv} > $frres->{pconv}) {
		$fres = $ffres;
		$fdir = 1;
		last;
	    }

	    $fres = $ffres;
	    $fdir = 1;
	}
    }

    unless ($conv == 0) {
	$rfres = execNeedle ($qfilepF, $gfilepR, $cpgr);
	$rrres = execNeedle ($qfilepR, $gfilepR, $cpgr);

	{
	    if ($rfres->{'aliMis'} > $rrres->{'aliMis'}) {
		$rres = $rrres;
		$rdir = -1;
		last;
	    }

	    if ($rfres->{'aliMis'} < $rrres->{'aliMis'}) {
		$rres = $rfres;
		$rdir = 1;
		last;
	    }

	    if ($rfres->{perc} > $rrres->{perc}) {
		$rres = $rfres;
		$rdir = 1;
		last;
	    }

	    if ($rfres->{perc} < $rrres->{perc}) {
		$rres = $rrres;
		$rdir = -1;
		last;
	    }

	    if ($rfres->{unconv} > $rrres->{unconv}) {
		$rres = $rrres;
		$rdir = -1;
		last;
	    }

	    if ($rfres->{unconv} < $rrres->{unconv}) {
		$rres = $rfres;
		$rdir = 1;
		last;
	    }

	    if ($rfres->{pconv} < $rrres->{pconv}) {
		$rres = $rrres;
		$rdir = -1;
		last;
	    }

	    if ($rfres->{pconv} > $rrres->{pconv}) {
		$rres = $rfres;
		$rdir = 1;
		last;
	    }

	    $rres = $rfres;
	    $rdir = 1;
	}
    }

    if ($conv == 0) {
	$res  = $fres;
	$dir  = $fdir;
	$gdir = 1;

    } elsif ($conv == 1) {
	$res = $rres;
	$dir  = $rdir;
	$gdir = -1;

    } else {
	{
	    if ($fres->{'aliMis'} > $rres->{'aliMis'}) {
		$res  = $rres;
		$dir  = $rdir;
		$gdir = -1;
		last;
	    }

	    if ($fres->{'aliMis'} < $rres->{'aliMis'}) {
		$res  = $fres;
		$dir  = $fdir;
		$gdir = 1;
		last;
	    }

	    if ($fres->{perc} > $rres->{perc}) {
		$res  = $fres;
		$dir  = $fdir;
		$gdir = 1;
		last;
	    }

	    if ($fres->{perc} < $rres->{perc}) {
		$res  = $rres;
		$dir  = $rdir;
		$gdir = -1;
		last;
	    }

	    if ($fres->{unconv} > $rres->{unconv}) {
		$res  = $rres;
		$dir  = $rdir;
		$gdir = -1;
		last;
	    }

	    if ($fres->{unconv} < $rres->{unconv}) {
		$res  = $fres;
		$dir  = $fdir;
		$gdir = 1;
		last;
	    }

	    if ($fres->{pconv} < $rres->{pconv}) {
		$res  = $rres;
		$dir  = $rdir;
		$gdir = -1;
		last;
	    }

	    if ($fres->{pconv} > $rres->{pconv}) {
		$res  = $fres;
		$dir  = $fdir;
		$gdir = 1;
		last;
	    }

	    $res  = $fres;
	    $dir  = $fdir;
	    $gdir = 1;
	}
    }

    unless ($gdir == 1) {
	$res->{qAli} = revComp $res->{qAli};
	$res->{gAli} = revComp $res->{gAli};
	$res->{val}  = reverse revComp $res->{val};
    }

    $ref = {fa => $fa, res => $res, dir => $dir, gdir => $gdir, exc => 0};
    $ref->{exc} = 1 if ($res->{unconv} > $unc);
    $ref->{exc} = 1 if ($res->{pconv}  < $pcon);
    $ref->{exc} = 1 if ($res->{aliMis} > $mis);
    $ref->{exc} = 1 if ($res->{perc}   < $perc);
    push @data, $ref;
}

unlink $gfilepF;
unlink $gfilepR;
unlink $qfilepF;
unlink $qfilepR;
unlink CPGMAT;

$cpgn = @pos;

unless ($form) {
    print join "\t", ('genome', $conv, $gseq, @pos.'', join ',', @pos);
# first base = 0
    print  "\n";
    foreach my $ref (@data) {
	print join "\t", ($ref->{fa}{pos}, $ref->{fa}{com}, $ref->{fa}{seq},
			  $ref->{res}{qAli}, $ref->{res}{gAli},	$ref->{res}{aliLen},
			  $ref->{res}{aliMis}, $ref->{res}{perc}, $ref->{res}{gap},
			  $ref->{res}{menum}, $ref->{res}{unconv}, $ref->{res}{conv},
			  $ref->{res}{pconv}, $ref->{res}{val},
			  $ref->{dir}, $ref->{gdir});
	print "\n";
    }

} elsif ($form == 1) {

    $lc = LINE;
    $gseq =~ s/(\w{1,$lc})/$1\n/g;
    $gseq = ">genome sequence\n$gseq";
    %cpg = map {$_ => 1} @pos;

    foreach $ref (sort {$a->{fa}{pos} <=> $b->{fa}{pos}} @data) {

	$qlen = length $ref->{fa}{seq};
	$ref->{fa}{seq} =~ s/(\w{1,$lc})/$1\n/g;
	$ref->{fa}{seq} = ">$ref->{fa}{com}\n$ref->{fa}{seq}";

	$alen  = length $ref->{res}{qAli};

	for ($start = 0; $start < $alen; $start++) {
	    $g = substr $ref->{res}{gAli}, $start, 1;
	    last unless ($g eq '-');
	}

	for ($stop = $alen - 1; $stop >= 0; $stop--) {
	    $g = substr $ref->{res}{gAli}, $stop, 1;
	    last unless ($g eq '-');
	}

	if ($ref->{dir} * $ref->{gdir} == 1) {
	    $qstart = $start;
	    $qstop  = $qlen - $alen + $stop ;

	} else {
	    $qstart = $alen  - $stop - 1;
	    $qstop  = $qlen - $start - 1;
	}

	$output = '';

	if ($ref->{gdir} == 1) {
	    ($bbase, $abase) = ('C', 'T');
	} else {
	    ($bbase, $abase) = ('G', 'A');
	}

	for ($j = $k = $l = 0, $i = $start; $i <= $stop; $i++) {

	    $j = $i - $start + 1;
	    $g = substr $ref->{res}{gAli}, $i, 1;
	    $q = substr $ref->{res}{qAli}, $i, 1;
	    $k++ unless ($g eq '-');
	    $l++ unless ($q eq '-');

	    if ($j % $lc == 1) {
		$k++ if ($g eq '-');
		$l++ if ($q eq '-');
		$gout = 'Genome'   .' 'x(3 + 6 - length($k))."$k  ";
		$ali  =             ' 'x(9 + 6 + 2);
		$qout = 'Bisulfite'.' 'x(0 + 6 - length($l))."$l  ";
		$k-- if ($g eq '-');
		$l-- if ($q eq '-');
	    }

	    $gout .= $g;
	    $qout .= $q;

	    if ($g eq $bbase) {

		$cpg = 0;

		if ($ref->{gdir} == 1 && $cpg{$k-1} ||
		    $ref->{gdir} != 1 && $cpg{$k-2}) {
		    $cpg = 1;
		}

		if ($q eq $bbase) {
		    if ($cpg) {
			$ali .= '*';
		    } else {
			$ali .= '!';
		    }
		} elsif ($q eq $abase) {
		    if ($cpg) {
			$ali .= '#';
		    } else {
			$ali .= ':';
		    }
		} else {
		    $ali .= ' ';
		}
	    } elsif ($g eq $q) {
		$ali .= '|';
	    } else {
		$ali .= ' ';
	    }

	    if ($j % $lc == 0) {
		$gout .= "\n";
		$ali  .= "\n";
		$qout .= "\n";

		$output .= $gout . $ali . $qout . "\n";
		$gout = $ali = $qout = '';
	    }
	}

	if ($j % $lc > 1) {
	    $gout .= "\n";
	    $ali  .= "\n";
	    $qout .= "\n";
	    $output .= $gout . $ali . $qout;
	    $gout = $ali = $qout = '';
	}

	$ref->{dir}  = $ref->{dir} * $ref->{gdir} == 1 ? 'forward' : 'reverse';
	$ref->{gdir} = $ref->{gdir} == 1 ? 'forward' : 'reverse';
	$cph  = $ref->{res}{unconv} + $ref->{res}{conv};
	$cpg  = $ref->{res}{val} =~ y/01//;
	$meth = $ref->{res}{val};
	$meth =~ y/MRWSYKVHDB/mrwsykvhdb/;
	$meth =~ y/01/UM/;
	$cpgh = $cpg ? 100 * $ref->{res}{menum}/$cpg : 0;

	printf ALI, $ref->{fa}{pos}, $ref->{fa}{com}, $bbase, $abase,
	$ref->{gdir}, $qlen, $len, $qstart + 1, $qstop + 1,
	$ref->{dir}, $cpg, $ref->{res}{menum}, $cpgh,
	$ref->{res}{unconv}, $cph, $ref->{res}{pconv},
	$ref->{res}{aliMis}, $ref->{res}{gap}, $ref->{res}{aliLen},
	$ref->{res}{perc}, $meth, $gseq, $ref->{fa}{seq}, $output;
    }

} elsif ($form == 2) {

    @data = @data[map {$_->{fa}{pos} - 1} grep {$_->{'exc'} == 0} @data];

    for ($i = 0; $i < $len; $i++) {
	$mali[$i]->{g} = substr $gseq, $i, 1;
    }

    for ($i = 0; $i < @data; $i++) {

	$ref = $data[$i];
	$alen = length $ref->{res}{gAli};

	for ($start = 0; $start < $alen; $start++) {
	    $g = substr $ref->{res}{gAli}, $start, 1;
	    last unless ($g eq '-');
	}

	for ($stop = $alen - 1; $stop >= 0; $stop--) {
	    $g = substr $ref->{res}{gAli}, $stop, 1;
	    last unless ($g eq '-');
	}

	for ($j = $k = 0; $j < $stop - $start + 1; $j++) {

	    $g = substr $ref->{res}{gAli}, $start + $j, 1;
	    $q = substr $ref->{res}{qAli}, $start + $j, 1;

	    if ($g eq '-') {
		$mali[$k]->{gap}{$i} .= $q;

	    } else {
		$g eq $mali[$k]->{g} or die "Genome sequence mismatch";
		$mali[$k]->{$i} = $q;
		$k++;
	    }
	}
    }

    for ($i = $j = 0; $i < @mali; $i++, $j++) {

	$alen = 0;

	if (exists $mali[$i]->{gap}) {
	    foreach $k (keys %{$mali[$i]->{gap}}) {
		$alen = length $mali[$i]->{gap}{$k}
	        if ($alen < length $mali[$i]->{gap}{$k});
	    }
	}

	$gen .= '-' x $alen . $mali[$i]->{g};

	for ($k = 0; $k < @data; $k++) {

	    $ref = $data[$k];

	    if ($alen) {
		$mali[$i]->{gap}{$k} ||= '';
		$mali[$i]->{gap}{$k} .=
		    '-' x ($alen - length ($mali[$i]->{gap}{$k}));
		$ref->{multi} .= $mali[$i]->{gap}{$k}
	    }

	    $ref->{multi} .= $mali[$i]->{$k};
	}
	$j += $alen;
    }


    $out = "genomic\t$gen\n";
    foreach $ref (@data) {
	$out .= "$ref->{fa}{com}\t$ref->{multi}\n";
    }

    print $out;

} else {
# show data

    for ($j = 1; $j <= $cpgn; $j++) {
	$meth{$j}{total} = 0;
	$meth{$j}{unmeth} = 0;
	$meth{$j}{meth} = 0;
	$meth{$j}{ng} = 0;
    }

    $allM = $allC = 0;
    $numT = @data;
    $numU = 0;
    $numE = 0;
    $out = '';
    $i = 0;

    foreach $ref (@data) {
	$meth = $ref->{gdir} == 1 ? reverse $ref->{res}{val} : $ref->{res}{val};
	$pat = '';

	if ($ref->{exc}) {
	    $numE++;

	} else {
	    $numU++;

	    for ($j = 1; $j <= $cpgn; $j++) {
		$m = chop $meth;

		if ($m eq '0') {
		    $meth{$j}{total}++;
		    $meth{$j}{unmeth}++;
		    $allC++;
		    $pat .= 'U';
		} elsif ($m eq '1') {
		    $meth{$j}{total}++;
		    $meth{$j}{meth}++;		    $allC++;
		    $allM++;
		    $pat .= 'M';
		} else {
		    $meth{$j}{ng}++;
		    $m =~ y/MRWSYKVHDB/mrwsykvhdb/;
		    $pat .= $m;
		}
	    }
	}

	$pat ||= 'excluded';
	$cpg  = $ref->{res}{val} =~ y/01//;
	$cpgp = $cpg ? 100 * $ref->{res}{menum} / $cpg : 0;
	push @dsds, $cpgp unless ($ref->{exc}); 
	$cpgp = sprintf "%5.1f", $cpgp;

	$out .= join "\t", (++$i, qq{$ref->{fa}{com}}, $ref->{res}{aliMis},
			    $ref->{res}{gap}, $ref->{res}{aliLen},
			    $ref->{res}{perc}, $ref->{res}{menum},
			    $cpgp, $ref->{res}{unconv},
			    $ref->{res}{unconv} + $ref->{res}{conv},
			    $ref->{res}{pconv}, $pat);
	$out .= "\n";
    }


    @dsdc = map {$_->{total} ? 100 * $_->{meth}/$_->{total} : 0}
            map {$meth{$_}} sort {$a <=> $b} keys %meth;

    $sds = @dsds ? stddev @dsds : 0;
    $sdc = @dsdc ? stddev @dsdc : 0;
    $ses = $sds / sqrt(@dsds || 1);
    $sec = $sdc / sqrt($cpgn || 1);


    $val1 = join "\t", @pos, 'total';
    $val2 = join ("\t", map {$_->{meth}}
		        map {$meth{$_}}
		        sort {$a <=> $b}
	                keys %meth);
    $val3 = join ("\t", map {$_->{total}}
	                map {$meth{$_}}
	                sort {$a <=> $b}
	                keys %meth);
    $val4 = join ("\t", map {sprintf "%.1f", $_} @dsdc);

    $val2 .= "\t$allM";
    $val3 .= "\t$allC";
    $val4 .= sprintf "\t%.1f", ($allC ? 100 * $allM / $allC : 0);

    printf OUTFORM, $len, $cpgn, $numU, $numE, $numT,
    $unc, $pcon, $mis, $perc,
    $val1, $val2, $val3, $val4,
    $sdc, $sec, $sds, $ses, $out;

}

###########################  End of Main ######################################
exit; #--1----+----2----+----3----+----4----+----5----+----6----+----7----+----
###########################  End of Main ######################################

sub usage () {
    print <<"EOS";
usage: $0 [options] - or input_file or -g genome_file -q query_file

Input data
    1) - (STDIN)
        Multi-FASTA format of genomic DNA sequence and bisulfite sequences
        First sequence must be genomic DNA seuqnce
    2) input_file 
        Multi-FASTA format of genomic DNA sequence and bisulfite sequences
        First sequence must be genomic DNA seuqnce
    3) -g genome_file -q query_file
        genome_file : FASTA format genomic DNA sequence
        query_file: Multi-FASTA format bisulfite sequences

Option
    -f: output format (0|1|2|3) default 0
        0: tab separated data
           first line : 'genome', condition of convert direction (see below),
                        genomic sequence, number of CpG, CpG position (first base = 0)
           other lines: No., sequence name, sequence, alignment data of this sequnece,
                        alignment data of genome sequence, alignment length,
                        number of alignmnet mismatch, percent identity of alignment,
                        number of alignment gap, number of methylated CpG,
                        number of bisulfite unconverted CpH (CpH: CpA, CpC, CpT),
                        number of bisulfite converted CpH,
                        percent of bisulfite convertion,
                        CpG methylation pattern (0: unmethylated, 1: methylated),
                        convert direction (1: forward, -1: reverse)
        1: human readable alignment data
        2: tab separated multiple alignment data
        3: tab separated summarized data
     -d: condition of convert direction of genomic sequence (0|1|2) default 0
        0: C -> T conversion
           PCR primer pair was designded for forward strand of the genomic sequence
        1: G -> A conversion
           PCR primer pair was designded for reverse strand of the genomic sequence
        2: both
           Search both direction of conversion and adopt more appropriate strand
     -u: upper limit of unconverted CpHs (integer, default 5)
         (CpH: CpA, CpC, CpT)
     -c: lower limit of percent converted CpHs (float, default 95.0)
     -m: upper limit of alignment mismatches (integer, default 10)
     -p: lower limit of percent identity (float, default 90.0)
     -u, -c, -m -p options are only affected output format 3 or 4
EOS
    exit;
}

sub makeTime (;$) {

    my $time = shift || time;

    my @time;

    @time = (localtime ($time))[reverse (0..5)];

    $time[1]++;
    $time[0] += 1900;

    sprintf("%04d%02d%02d%02d%02d%02d", @time);
}

sub curateSeq ($) {

    my $seq = shift;

    $seq =~
	y/ACGTURYMWSKDHBVNacgturymwskdhbvn//cd;
    $seq;
}

sub parseSeq ($) {

    my ($seq ) = shift;
    my ($com );
    local ($1);

    $seq or return;
    $com = '';
    $seq =~ s/\r\n/\n/g;
    $seq =~ s/\r/\n/g;
    $seq = uc $seq;

    if ($seq =~ /^\s*>.*?\n/m) {
	$seq =~ s/^\s*>(.*)?\n//m;
	$com = $1;
	$com =~ s/\t//g;
	$com =~ s/\s*$//g;

    } elsif ($seq =~ /^ORIGIN\s*\n((\s+(\d+(\s+\w+)+))+)\s*\n\/\//m) {
	$seq = $1;

    } elsif ($seq =~ /^SQ\s+SEQUENCE.*\n((\s+(\w+\s+)+\d+)+)\n\/\//m) {
	$seq = $1;

    } elsif ($seq =~ /\.\.\s*\n((\s+(\d+(\s+\w+)+))+)\s*/m) {
	$seq = $1;
    } elsif ($seq =~ /^\s*>.+\s.+/m) {
	$seq =~ s/^\s*>(.+?)\s(?=.+)//m;
	$com = $1;
    }

    return (curateSeq ($seq));
}

sub parseGenome ($) {

    my ($file) = shift;
    my ($seq );
    my ($fh  );
    local ($!);
    local ($1);

    open $fh, $file or die "$!:$file";
    $seq = join '', <$fh>;
    close $fh;

    $seq =~ s/^[\r\s]+//;
    $seq =~ s/[\r\s]+$//;
    $seq =~ s/(\r|\n|\r\n){2}/$1/g;

    return (parseSeq ($seq));
}

sub parseBiseq ($) {

    my ($file ) = shift;
    my ($multi);
    my ($fh   );

    open $fh, $file or die "$!:$file";
    $multi = join '', <$fh>;
    close $fh;

    $multi =~ s/^[\r\s]+//;
    $multi =~ s/[\r\s]+$//;
    $multi =~ s/(\r|\n|\r\n){2}/$1/g;

    return (multiFastaParse ($multi));
}


sub parseMulti ($) {

    my ($file ) = shift;
    my ($multi);
    my ($fh   );
    my (@biseq);
    my ($seq  );

    open $fh, $file or die "$!:$file";
    $multi = join '', <$fh>;
    close $fh;

    $multi =~ s/^[\r\s]+//;
    $multi =~ s/[\r\s]+$//;
    $multi =~ s/(\r|\n|\r\n){2}/$1/g;
    
    ($seq, @biseq) = multiFastaParse ($multi);
    return ($seq->{seq}, @biseq);
}

sub multiFastaParse ($) {

    my ($multi) = shift;
    my ($fa   );
    my ($line );
    my (@biseq);
    local ($1 );

    $multi =~ s/\r\n/\n/g;
    $multi =~ s/\r/\n/g;

    while ($multi =~ /(.*)$/mg) {

	chomp ($line = $1);

	if ($line =~ /^>/) {
	    pop @biseq if ($fa && ! $fa->{seq});
	    $fa = {};
	    push @biseq, $fa;
	    $fa->{com} = $line;
	    $fa->{com} =~ s/^>//;
	    $fa->{com} =~ s/\t//g;
	    $fa->{com} =~ s/\s*$//g;
	} else {
	    $line = curateSeq ($line);
	    $line or next;
	    $fa	or return ();
	    $fa->{seq} .= uc $line; 
	}
    }

    pop @biseq if ($fa && ! $fa->{seq});
    return (@biseq);
}

sub revComp ($) {

    my $seq = reverse shift;

    $seq =~
	y/ACGTURYMWSKDHBVNacgturymwskdhbvn/TGCAAYRKWSMHDVBNtgcaayrkwsmhdvbn/;
    $seq;
}

sub fastaMake ($$;$) {

    my $seq  = shift;
    my $com  = shift;
    my $line = shift || 60;

    local $1;

    $seq =~ y/0-9 \t\n\r\f//d;

    $seq =~ s/(.{1,$line})/$1\n/g;

    ">$com\n" . $seq;
}

sub fastaPrint ($$$;$$) {

    my $seq  = shift;
    my $com  = shift;
    my $path = shift;
    my $line = shift;
    my $add  = shift;
    my $fh;

    local ($!  );

    if ($add) {
	open ($fh, ">>$path") or die "$!:$path";

    } else {
	open ($fh, ">$path") or die "$!:$path";
    }

    print $fh fastaMake ($seq, $com, $line);
    close ($fh);
}

sub execNeedle ($$$) {

    my ($gfile) = shift;
    my ($qfile) = shift;
    my ($cpg  ) = shift;
    my ($com  );
    my ($fh   );
    my ($ref  );
    my ($gAli );
    my ($qAli );
    my ($i    );
    my ($flag );
    my ($g    );
    my ($q    );
    my ($j    );
    my ($pos  );
    local ($! );
    local ($_ );
    local ($? );

    $ref = {};
    $ref->{qAli}   = '';
    $ref->{gAli}   = '';
    $ref->{gap}    = 0;
    $ref->{menum}  = 0;
    $ref->{unconv} = 0;
    $ref->{conv}   = 0;
    $ref->{pconv}  = '';
    $ref->{match}  = 0;
    $ref->{val}    = '';
    $ref->{perc}   = '';
    $ref->{aliMis} = 0;

    $com = NEEDLE . " $gfile $qfile " . NDLOPT;

    open $fh, $com or die "$!:$com";

    while (<$fh>) {

	chomp;

	if (/^>que/../^>genome/) {
	    $ref->{qAli} .= $_ unless (/^>/);
	}
	if (/^>genome/../^$/) {
	    $ref->{gAli} .= $_ unless (/^>/);
	}
    }

    $ref->{qAli} =~ y/ /-/;
    $ref->{gAli} =~ y/ /-/;

    close $fh or die "$?:$com";

    length $ref->{gAli} == length $ref->{qAli} or
	die ("Needle error! gAli len != qAli len : " .
	     (length $ref->{gAli}) . " !=  " . (length $ref->{qAli}) . ";;;" .
	     $ref->{gAli} . "::::" . $ref->{qAli});

    $gAli = $qAli = '';

    for ($i = $flag = 0; $i < length $ref->{gAli}; $i++) {

	$g = uc substr $ref->{gAli}, $i, 1;
	next if ($g eq '-' && $flag == 0);
	$flag = 1;
	$q = uc substr $ref->{qAli}, $i, 1;
	$gAli .= $g;
	$qAli .= $q;
    }

    while ($g = chop $gAli) {
	unless ($g eq '-') {
	    $gAli .= $g;
	    last;
	}
	chop $qAli;
    }

    $ref->{aliLen} = length $qAli;

    for ($i = $j = 0; $i < $ref->{aliLen}; $i++) {

	$g = uc substr $gAli, $i, 1;
	$q = uc substr $qAli, $i, 1;

	$j++ unless ($g eq '-');

	if ($q eq $g ||
	    $g eq 'C' && $q eq 'T') {
	    $ref->{match}++;
	}

	if ($g eq '-') {
	    $ref->{gap}++;
	    next;
	}

	if ($q eq '-') {
	    $ref->{gap}++;
	    if ($cpg->{$j - 1}) {
		$ref->{val} .= '-';
	    }
	    next;
	}

	last if ($i == $ref->{aliLen} - 1);

	next unless ($g eq 'C');

	unless ($cpg->{$j - 1}) {
	    if ($q eq 'C') {
		$ref->{unconv}++;
	    } elsif ($q eq 'T') {
		$ref->{conv}++;
	    }
	    next;
	}

	if ($q eq 'C') {
	    $ref->{menum}++;
	    $ref->{val} .= '1';

	} elsif ($q eq 'T') {
	    $ref->{val} .= '0';

	} else {
	    $ref->{val} .= $q;
	}
    }

    if ($ref->{conv} + $ref->{unconv}) {
        $ref->{pconv} = sprintf "%3.1f", 100 * $ref->{conv}/($ref->{conv} + $ref->{unconv});
    } else {
	$ref->{pconv} = 0;
    }
    $ref->{perc}   = sprintf "%3.1f", 100 * $ref->{match} / $ref->{aliLen};
    $ref->{aliMis} = $ref->{aliLen} - $ref->{match};
    return $ref;
}

#---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----
###############################################################################
1;
###############################################################################
